//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#if !defined(__THQCPPbase_H__)
#define __THQCPPbase_H__ 1

#include <sys/types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
// #include <math.h>
#include <string.h>
#include <signal.h>

#include <vector>
#include <list>
#include <map>
#include <deque>
#include <queue>
#include <stack>
#include <chrono>
#include <thread>
#include <atomic>
#include <mutex>
#include <condition_variable>

using std::queue;
using std::vector;
using std::list;
using std::map;
using std::stack;
using std::deque;
using std::mutex;
using std::unique_lock;
using std::condition_variable;

//#include "autop.h"

#define THQALIGNED __attribute__((aligned(64)))

#define THQINLINE __attribute__((always_inline))

#define THQNOINLINE __attribute__((noinline))

#if !defined(THQFATTR0)
#define THQFATTR0 __attribute__((noinline))
#endif
#if !defined(THQFATTR1)
#define THQFATTR1 __attribute__((noinline))
#endif
#if !defined(THQFATTR2)
#define THQFATTR2 __attribute__((noinline))
#endif

#define THQCPPVER

#include "THQdefs.h"

//
// Inter thread buffering, uses pthread mutex/cond variable based synchronization.
//

enum class THQpayloadstate {
	STNOINFO = 0,
	STEOF = 1
};

//
// Data sent between threads "class THQpayload"
// Data communications mgmt is handled by "class Tachunkqueue" and
//   "class THQstack".  Tachunkclass provides more immediate
//   reuse of data chunks -- POSSIBLY providing better locality of
//   reference.  I found that if the system is very busy with the
//   program, then the reuse because less likely.  It seems to be
//   a good idea to try to implement the more recent reuse if possible,
//   it just might help and doesn't add anything to the complexity.
//

//
// This is the raw data that is sent between threads
//
template <typename ASTYPE>
class THQALIGNED THQpayload
{
	int index;
	THQpayloadstate state;
	ASTYPE THQALIGNED payloadbuf;

public:
	THQpayload() THQFATTR0 :
		index(0), state(THQpayloadstate::STNOINFO), payloadbuf() {}

	THQpayload(const THQpayload<ASTYPE>& init) THQFATTR0 :
		index(init.index), state(THQpayloadstate::STNOINFO),
		payloadbuf(init.payloadbuf)
	{
	}

	~THQpayload() THQFATTR0
	{
	}

	THQpayload& THQFATTR0
	operator=(const THQpayload<ASTYPE> & src)
	{
		if (&src == this)
			return *this;

		index = src.index;
		state = src.state;
		payloadbuf = src.payloadbuf;
		return *this;
	}

	//
	// Each raw data buffer has an 'index' or 'id' for logging/statistics purposes.
	//
	int 
	getindex(void) const
	{
		return index;
	}

	void 
	setindex(int newindex)
	{
		index = newindex;
	}

	//
	// This supports access to the raw data in the buffer.
	//
	const ASTYPE&
	payload(void) const
	{
		return payloadbuf;
	}

	ASTYPE&
	payload(void)
	{
		return payloadbuf;
	}

	bool
	iseof(void) const
	{
		return state == THQpayloadstate::STEOF;
	}

	void
	seteof(void)
	{
		state = THQpayloadstate::STEOF;
	}

	void
	flush(void)
	{
#if 1
		char *pb = (char *) &payloadbuf;

		for (size_t i = 0; i < sizeof payloadbuf; i += 64)
		{
			asm volatile ("clflush %0" : : "m"(pb[i]));
		}
#endif
	}
};

template <class ELEMENT>
	class THQstack {

	protected:
		std::deque<ELEMENT> c;
	public:

		THQstack(void) THQFATTR0 : c() {}

		THQstack(const THQstack& init) THQFATTR0 : c(init.c) {}


		typename std::deque<ELEMENT>::size_type THQFATTR0 size() const
		{
			return c.size();
		}

		bool THQFATTR0 empty(void) const
		{
			return c.empty();
		}

		void THQFATTR0 additem(const ELEMENT& e)
		{
			c.push_back(e);
		}

		ELEMENT THQFATTR0 getnextitem(void)
		{
			//
			// Somehow check for error -- lets avoid exceptions
			//
			ELEMENT e(c.back());
			c.pop_back();
			return e;
		}

		ELEMENT& THQFATTR0 checknextitem(void)
		{
			return c.back();
		}

		static bool flushtarget(void)
		{
			return false;
		}
	};

template <class ELEMENT>
	class THQqueue {
	
	protected:
		std::deque<ELEMENT> c;

	public:
		THQqueue(void) THQFATTR0 : c() {}

		THQqueue(const THQqueue& init) THQFATTR0 : c(init.c) {}

		typename std::deque<ELEMENT>::size_type THQFATTR0 size() const
		{
			return c.size();
		}

		bool THQFATTR0 empty() const
		{
			return c.empty();
		}

		void THQFATTR0 additem(const ELEMENT& e)
		{
			c.push_back(e);
		}

		ELEMENT THQFATTR0 getnextitem(void)
		{
			ELEMENT e(c.front());
			c.pop_front();
			return e;
		}

		ELEMENT& THQFATTR0 checknextitem(void)
		{
			return c.front();
		}

		static bool flushtarget(void)
		{
			return true;
		}
	};


template <typename QTYPE, typename DTYPE, int MAXQSIZE, int QWAKEUPSIZE, bool NOBLOCKSTORE=false>
class THQ
{
	int curqcount;
	THQALIGNED QTYPE aq;
	mutex qlock;
	condition_variable waitdataavail;
	condition_variable waitqueueavail;
	char *qname;
	int64_t waitdataavailcount;
	int64_t signaldataavailcount;
	int64_t waitqueueavailcount;
	int64_t signalqueueavailcount;
	int64_t signalqueueavailoptimize;
	volatile int waitingondataavail;
	volatile int waitingonqueuespaceavail;

#if defined(DEFAULTWAKEUP)
	void NOINLINE
	local_pthread_cond_timewait50msec(pthread_cond_t* cond, pthread_mutex_t* mtx)
	{
		timespec waittimeout = THQsupport::gettime_with_offset_ms(50);
		pthread_cond_timedwait(cond, mtx, &waittimeout);
	}
#endif

#if defined(PRINTFTRACE)
	void NOINLINE
	waitqmsgs(int msgid)
	{
		switch (msgid)
		{
		case 0:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" is empty, waiting on data available\n", qname);
			break;
		case 1:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" error -- empty queue with curqcount == 0\n", qname);
			break;
		case 2:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" broadcast to cond variable waitqavail\n", qname);
			break;
		case 3:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" broadcast to cond variable waitqavail\n", qname);
			break;
		default:
			break;
		}
	}

	void NOINLINE
	waitqrundownmsg(int msgid, int msgarg = 0)
	{
		switch (msgid)
		{
		case 0:
			::fprintf(stderr,"THQwqrundown: queue \"%s\" rundown attempt %d\n", qname, msgarg);
			break;
		case 1:
			::fprintf(stderr,"THQwqrundown: queue \"%s\" rundown complete\n", qname);
			break;
		default:
			break;
		}
	}

	void NOINLINE
	waitqspacemsg(int msgid, bool foreof = false)
	{
		switch(msgid)
		{
		case 0:
			::fprintf(stderr,"THQwqspace: queue \"%s\" waiting for some space%s -- curqcount == %d\n",
				qname, foreof ? " for EOF processing" : "", curqcount);
			break;
		case 1:
			::fprintf(stderr,"THQwqspace: queue \"%s\" space found%s -- curqcount == %d\n",
				qname, foreof ? " for EOF processing" : "", curqcount);
			break;
		default:
			break;
		}
	}

	void NOINLINE
	settingnewqname(int msgid, const char *qnargorig, const char *qnargnew)
	{
		switch(msgid)
		{
		case 0:
			::fprintf(stderr,"setnewqname: original name: %s, newname: %s\n",
				(qnargorig != nullptr) ? qnargorig : "<unknown or undef>",
				(qnargnew != nullptr) ? qnargnew : "<set to null name>");
			break;
		default:
			break;
		}
	}

#endif

	THQ(const THQ& init);
	THQ& operator=(const THQ& src);
			
public:
	THQ(const char *qn = nullptr) THQFATTR1 : curqcount(0), aq(),
		qlock(), waitdataavail(),
		waitqueueavail(), qname((qn == nullptr) ? nullptr: (::strdup(qn))),
		waitdataavailcount(0), signaldataavailcount(0), waitqueueavailcount(0), signalqueueavailcount(0),
		signalqueueavailoptimize(0),
		waitingondataavail(false), waitingonqueuespaceavail(false)
	{
	}

	~THQ() THQFATTR0
	{
		if (qname != nullptr)
		{
			::free(qname);
			qname = nullptr;
		}
	}

	void THQFATTR1
	setqname(const char* newqname)
	{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 4)
		settingnewqname(0, qname, newqname);
#endif
		if (qname != nullptr)
		{
			::free(qname);
			qname = nullptr;
		}

		if (newqname != nullptr)
		{
			qname = ::strdup(newqname);
		}
	}


	DTYPE THQFATTR1
	THQwqdataread(void)
	{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 8)
		bool didwait = false;
#endif
		unique_lock<mutex> lck(qlock);

		while (aq.empty())
		{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 8)
			didwait = true;
#endif
			atomicstore(waitingondataavail, true);
			++waitdataavailcount;
#if defined(PRINTFTRACE) && (PRINTFTRACE > 8)
			waitqmsgs(0);
#endif

#if defined(DEFAULTWAKEUP)
			waitdataavail.wait_for(lck, std::chrono::milliseconds(50));
#else
			waitdataavail.wait(lck);
#endif
		}
#if defined(PRINTFTRACE) && (PRINTFTRACE > 8)
		if (didwait)
			::fprintf(stderr,"THQwqdataread***: queue \"%s\" data avail after WAITING -- curqcount == %d\n",
				qname, curqcount);
#endif

		DTYPE dqentry = aq.getnextitem();

		atomicaddload(curqcount, -1);

		atomicstore(waitingondataavail, false);

		//
		// Count the freed up queue space
		// and wakeup anyone who is waiting.
		//
		// We send signal if we need more.
		//
		if (atomicload(curqcount) <= QWAKEUPSIZE)
		{
			if (atomicload(waitingonqueuespaceavail))
			{
				waitqueueavail.notify_one();
				++signalqueueavailcount;
			}
			else
			{
				++signalqueueavailoptimize;
			}
		}

#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
		::fprintf(stderr,"THQwqdataread***: queue \"%s\" data received -- curqcount == %d\n",
			qname, curqcount);
#endif

		return dqentry;
	}

	//
	// This supports queue filling at startup
	//
	int THQFATTR1
	THQwakeupsource(void)
	{
		int retrycount = 10;
		while ((retrycount > 0) && (atomicload(curqcount) <= QWAKEUPSIZE))
		{
//			pthread_yield();
			std::this_thread::yield();
			if (!atomicload(waitingonqueuespaceavail))
			{
				break;
			}

			waitqueueavail.notify_one();
			++signalqueueavailcount;
		}

		return curqcount;
	}

	void THQFATTR1
	THQwqrundown(void)
	{
		//
		// since the purposeof this routine is to rundown the queue -- basically
		// when the system might not be totally sane, we try to rundown the queue
		// for a limited number of times/amount of time then we give-up.
		//
		int rundowncount = 10;
		int timeoutmult = 1;

		unique_lock<mutex> lck(qlock);

		for (int ntries = 0; ntries < rundowncount; ntries++)
		{
			{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
				waitqrundownmsg(0, ntries);
#endif
				waitdataavail.wait_for(lck, std::chrono::milliseconds(100 * rundowncount));
				rundowncount++;
				timeoutmult *= 2;
				if (timeoutmult > 100)
					timeoutmult = 100;
			}
			if (aq.empty())
			{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
				waitqrundownmsg(1);
#endif
				break;
			}
		}
	}

	void THQFATTR1
	THQwqspaceandinsert(DTYPE dqentry, bool deferwakeup = false)
	{
		unique_lock<mutex> lck(qlock);

		if (!NOBLOCKSTORE)
		{
			while (atomicload(curqcount) >= MAXQSIZE)
			{
				++waitqueueavailcount;
				atomicstore(waitingonqueuespaceavail, true);
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
				waitqspacemsg(0, dqentry->iseof());
#endif
#if defined(DEFAULTWAKEUP)
				waitqueueavail.wait_for(lck, std::chrono::milliseconds(50));
#else
				waitqueueavail.wait(lck);
#endif
			}
		}

		//
		// Have to save iseof here, because we don't theoretically own dqentry once
		// we have queued it...  Yeah, I know that we haven't released the mutex yet,
		// but trying to be as clean as possible.
		//
		bool iseof = dqentry->iseof();
		atomicstore(waitingonqueuespaceavail, false);
		atomicaddload(curqcount, 1);
		aq.additem(dqentry);
		if (aq.flushtarget())
			dqentry->flush();

		//
		// Wakeup if we don't ask for deferral or if the queue is full
		//
		bool donotify = !deferwakeup || (atomicload(curqcount) >= MAXQSIZE) || iseof;

		if (donotify && atomicload(waitingondataavail))
		{
			waitdataavail.notify_one();
			++signaldataavailcount;
		}

#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
		waitqspacemsg(1, iseof);
#endif
		return;
	}

	void THQFATTR1
	THQwqspaceandeof(DTYPE dqentry)
	{
		dqentry->seteof();
		THQwqspaceandinsert(dqentry);
		return;
	}

	void NOINLINE
	dumpqueuestatus(void)
	{
#if 0
		::fprintf(stderr,"queue %s: waitd: %lld, signald: %lld, waitq: %lld, signalq: %lld, signalqopt: %lld\n",
			qname, (long long) waitdataavailcount, (long long) signaldataavailcount, (long long) waitqueueavailcount,
			(long long) signalqueueavailcount, (long long) signalqueueavailoptimize);
#endif
	}

	int
	THQqcount(void) const
	{
		return curqcount;
	}

};

#endif

//
// vim: set ts=4:
//
